;**********************************************************************
;                                                                     *
;    Filename:	    mrx-008.asm                                       *
;    Date:                                                            *
;    File Version:                                                    *
;                                                                     *
;    Author:        el@jap.hu                                         *
;                   http://jap.hu/                                    *
;**********************************************************************
;NOTES
;
; Manchester decoding routine
;
; supposed crystal freq is 4MHz
; T ~= 620 usec (half bit)
;
;**********************************************************************
;HISTORY
;
;000-20010108
;initial version, for MPSIM debugging - simple bit decoding tested
; - behaviour on errors is not tested
;
;001-20010108
;to be linked with other external sources
;
;002-20010109
;protocol decoding tested
;
; header = 16xbit1, 1xbit0 (min/max n. of bit1s in header reqs are not yet checked)
; byte = 8xbitX, 1xbit1, 1xbit0
;
; if not in header, max 9xbit1 or 9xbit0 can be received (which can't be sync'ed to)
;
;003-20010110
;radio receiving tested OK - debug info on LCD
; counting 0-4 encoding errors/rcvd packets, displaying protocol errors
;
;004-20010113
;mrx algorythm without using TMR0
; (whis is reserved for measuring timeouts)
;
;005-20010113
;debug on LED. 0=OK, other=error. Point shows validity
;
;006-20010113
;same as 005, plus check buffer checksum and devid
;
;007-20010113
;raw OUT, 5-byte buffer, adapt to C508
;
;008-20010121
;based on 007 and 004, monitor version with LCD
;PB7 enters monitor mode without devid check
;monitor mode shows received valid packet data
;
;**********************************************************************
#define F84
;define C508

#ifdef F84
	list      p=16F84a             ; list directive to define processor

	__CONFIG   _CP_OFF & _WDT_OFF & _PWRTE_ON & _XT_OSC
	;WDT=watchdog, PWRTE=power on timer, XT= crystal osc

#include <p16F84a.inc>         ; processor specific variable definitions
#include "lcdlib.inc"
#define RXBIT PORTA, 0
#define CTLPORT PORTB
#endif

#ifdef C508
	list	p=12c508
	#include <p12C508.inc>

	__CONFIG   _CP_OFF & _WDT_OFF & _MCLRE_OFF & _IntRC_OSC

#define RXBIT GPIO, 3
#define CTLPORT GPIO
#endif

;***** VARIABLES
#define IF_SHORT flags, 0
#define FIRST_HALF flags, 1
#define HEADER flags, 2
#define VALID flags, 7

if_short_val	EQU	1 ; bit value of IF_SHORT flag
first_half_val	EQU	2 ; bit value of FIRST_HALF flag

#ifdef F84
.mainda UDATA 0x10
crsrpos		res	1
counters	res	8 ; error counters 0-5
#endif

#ifdef C508
.mainda UDATA 0x07
#endif

bitcnt		res	1
tmrval		res	1 ; timer value
bt		res	1 ; receive byte buffer
flags		res	1 ; decode logic status
btcnt		res	1 ; byte counter
agetmr		res	1 ; age out timer
buffer		res	4 ; receive packet buffer
bsum		res	1 ; receive buffer, checksum


;***** CONSTANT VALUES
hdrcntmin	EQU	0x0c
hdrcntmax	EQU	0x10

min_t		EQU	.34
min_2t		EQU	.104
max_2t		EQU	.174

age_max		EQU	.15 ; age out last valid packet
packet_len	EQU	5 ; packet length, check var. alloc!
devid1		EQU	0x9a
devid2		EQU	0xe9

#ifdef F84
vectors		CODE 0
  		goto    main		; go to beginning of program
		nop
		nop
		nop
					; interrupt vector location
		retfie

prog		CODE 5

	GLOBAL		messages
messages	movwf PCL
msg_line1	dt "1## 2## 3## 4## ", 0
msg_line2	dt "0## ------------", 0
msg_line3	dt "MrxV4 monitor   ", 0
msg_line4	dt "id #### ## ## ##", 0

position	addwf PCL
		retlw 0xc1 ; 1
		retlw 0x81 ; 2
		retlw 0x85 ; 3
		retlw 0x89 ; 4
		retlw 0x8d ; 0
		retlw 0xc5 ; 5

#endif

#ifdef C508
		CODE 0
		movwf OSCCAL
		bsf STATUS, 5
		goto hwinit_508
#endif

mp_receive	; receive a full manchester-encoded packet

s3		; set flags: first_half=1, if_short=0
		bsf FIRST_HALF
s4		bcf IF_SHORT

s5		; init before the received packet

		; set FSR to buffer start
		movlw buffer
		movwf FSR
		; set byte counter
		movlw packet_len ; bytes / packet
		movwf btcnt
		; set header receive mode
		bsf HEADER
		clrf bitcnt ; counting bit1-s in this mode

s2		; wait for a pulse
		btfss RXBIT
		goto s2

s6		; wait for end of (short) pulse up to min_2t
		clrf tmrval
s6_w		btfss RXBIT
		goto s7 ; goto s7 at end of pulse

		incf tmrval, F
		nop
		movlw min_2t
		subwf tmrval, W
		btfss STATUS, C
		goto s6_w

		; timeout, exit
		retlw 1 ; illegal startbit

s7		; start timer
		clrf tmrval

s8		; if (if_short & rxbit) goto s9
		; if (!if_short & !rxbit) goto s9
		; goto s10

		btfsc IF_SHORT
		; if_short = 1
		goto s8_ss1

s8_ss0		; if_short = 0
		btfsc RXBIT
		goto s10 ; rxbit = 1, goto s10

s9_ss0		; if (timer > max_2t) exit - else goto s8
		movlw max_2t
		subwf tmrval, W
		btfsc STATUS, C
		retlw 2 ; signal too long

		incf tmrval, F
		goto s8_ss0

s8_ss1		; if_short = 1
		btfss RXBIT
		goto s10 ; rxbit = 0, goto s10

s9_ss1		; if (timer > max_2t) exit - else goto s8
		movlw max_2t
		subwf tmrval, W
		btfsc STATUS, C
		retlw 2 ; signal too long

		incf tmrval, F
		goto s8_ss1

s10		; invert if_short
		movlw if_short_val
		xorwf flags, F

s11		; if (timer < min_t) exit
		movlw min_t
		subwf tmrval, W
		btfss STATUS, C
		retlw 3 ; signal too short

s12		; if (timer < min_2t) goto s14
		movlw min_2t
		subwf tmrval, W
		btfss STATUS, C
		goto s14

s13		; if (first_half = 0) goto s16 - else exit
		btfss FIRST_HALF
		goto s16
		retlw 4 ; no mid-frame transition/out of sync

s14		; invert first_half
		movlw first_half_val
		xorwf flags, F

s15		; if (first_half = 1) goto 7
		btfsc FIRST_HALF
		goto s7

s16		; if_short is a decoded bit. Handle here
		btfss HEADER
		goto s16_not_header

		; header receiving mode
		btfss IF_SHORT
		goto s16_header_end

		; header bit is 1
		btfss bitcnt, 4 ; inc up to 16
		incf bitcnt, F  ; 16 is enough...

		call age ; age out even if never leaving this loop
#ifdef NOMAXHDR
		; test for max header length
		movlw hdrcntmax
		subwf bitcnt, W
		btfss STATUS, C
#endif
		goto s7 ; loop back
		retlw 9 ; header too long

s16_header_end	; header ends indicated by a 0
		bcf HEADER

		; test for min header length
		movlw hdrcntmin
		subwf bitcnt, W
		btfss STATUS, C
		retlw 0x0a ; header too short

next_byte	movlw 0x0a
		movwf bitcnt
		goto s7 ; loop back

s16_not_header	; receiving bytes
		decf bitcnt, F
		bz s16_s4 ; if (bitcnt = 0) check for a byte-sep 1

		; if (bitcnt = 1) check for a byte-separator 0
		movlw 1
		xorwf bitcnt, W
		bnz s16_s2

		; test for a byte separator 1
		btfsc IF_SHORT
		goto s7
		retlw 7 ; byte-ending 1 not present

s16_s2		; bit is data

		rrf flags, W
		rlf bt, F
		goto s7

s16_s4		; check for a byte-separator 0
		btfsc IF_SHORT
		retlw 8 ; byte-ending 0 not present

		; OK, received byte is sane, store in buffer
		movf bt, W
		movwf INDF
		incf FSR, F

		decfsz btcnt, F
		goto next_byte

		retlw 0 ; OK, buffer received

chk_buf		; check buffer sanity by chksum
		movlw buffer
		movwf FSR
		movlw (packet_len-1) ; number of bytes w/o the chksum byte
		movwf btcnt
		clrw

chk0		addwf INDF, W
		incf FSR, F
		decfsz btcnt, F
		goto chk0
		;chksum is now in W

		subwf bsum, W
		bnz chk_err
		retlw 0 ; result is in Z
chk_err		retlw 0x0c ; checksum error

chk_devid	;check if the device id is the same
		; (first two bytes)

		movlw devid1
		subwf buffer, W
		bnz chk_err2

		movlw devid2
		subwf (buffer+1), W
		bnz chk_err2

		retlw 0 ; OK
chk_err2	retlw 0x0d ; devid error

#ifdef C508
hwinit_508	;C508 HARDWARE INIT
		clrf GPIO
		clrw; movlw 0x08
		tris GPIO ; all except GP3 is output
		clrwdt
		movlw 0x47 ; no pullup on GP3 (1&0 outputs)
		option
#endif

main
#ifdef F84
		;F84 HARDWARE INIT
		MOVLW 0 ;B00000000 ;SET LEVELS
		MOVWF PORTA
		MOVWF PORTB

		movlw 1 ; PORTA0 is input
		TRIS PORTA

		MOVLW 0x80 ; PORTB for LCD
		TRIS PORTB

		clrwdt ; changing default presc. assignment
		movlw 0x07 ; prescaler 1:256 assigned to TMR0
		OPTION     ; T0CS selects internal CLK
#endif

		clrf TMR0
		call age_clr ; no valid packet

		call lcd_init
warm_1		movlw 0x80
		call lcd_cmdout
		movlw msg_line1
		call lcd_strout
		movlw 0xc0
		call lcd_cmdout
		movlw msg_line2
		call lcd_strout

wa_1		btfss PORTB, 7
		goto wa_1

		call age_clr ; no valid packet

crhome		movlw 0xc4
		movwf crsrpos

		movlw 0x0e ; DispOn, CursOn, BlinkOff
		call lcd_cmdout

loop		btfss PORTB, 7
		goto warm_2

		movlw 0xcf
		call lcd_cmdout

		call age
		call mp_receive
		andlw 0xff
		bnz l_skip
		call chk_buf
		andlw 0xff
		bnz l_skip
		call chk_devid

l_skip		movwf bt ; 0=OK, other=error

		iorlw 0x30
		call lcd_chrout

		movf bt, W
		btfsc STATUS, Z
		call age_set

		movlw 6
		subwf bt, W
		btfsc STATUS, C
		goto not_enc_err

		; set LCD position according to code
		movf bt, W
		call position
		call lcd_cmdout

		movlw counters ; error code counter
		addwf bt, W
		movwf FSR
		incf INDF, F

		movf INDF, W
		call lcd_hexout
		goto loop

not_enc_err	; error codes 6 and up

		movf crsrpos, W
		call lcd_cmdout

		movf bt, W
		iorlw 0x30
		call lcd_chrout
		incf crsrpos, F
		movlw 0xcf
		subwf crsrpos, W
		bnc loop
		goto crhome

age		; time out last packet
		btfss TMR0, 7
		return

		clrf TMR0
		decfsz agetmr, F
		return

age_clr		; age out
		bcf VALID
		movlw age_max
		movwf agetmr

		;clrf CTLPORT ; clear all outputs
		movlw 0x8f
		call lcd_cmdout
		movlw '-'
		call lcd_chrout
		return

age_set		; validate now
		bsf VALID
		movlw age_max
		movwf agetmr

		; show validity here
		movlw 0x8f
		call lcd_cmdout
		movlw 'V'
		call lcd_chrout
		return

warm_2		movlw 0x80
		call lcd_cmdout
		movlw msg_line3
		call lcd_strout
		movlw 0xc0
		call lcd_cmdout
		movlw msg_line4
		call lcd_strout

wa_2		btfss PORTB, 7
		goto wa_2

		call age_clr ; no valid packet


loop_2		btfss PORTB, 7
		goto warm_1

		movlw 0x8e
		call lcd_cmdout

		call age
		call mp_receive
		andlw 0xff
		bnz l_skip_2
		call chk_buf
		andlw 0xff
		bnz l_skip_2
		;call chk_devid
		; no id check in monitor mode...

l_skip_2	movwf bt ; 0=OK, other=error

		iorlw 0x30
		call lcd_chrout

		movf bt, W
		btfss STATUS, Z
		goto loop_2

		call age_set
		movlw 0xc3
		call lcd_cmdout
		movf buffer, W
		call lcd_hexout
		movf buffer+1, W
		call lcd_hexout
		movlw ' '
		call lcd_chrout
		movf buffer+2, W
		call lcd_hexout
		movlw ' '
		call lcd_chrout
		movf buffer+3, W
		call lcd_hexout
		movlw ' '
		call lcd_chrout
		movf buffer+4, W
		call lcd_hexout

		goto loop_2



		end

